@@ -19,6 +19,14 @@
  * Boston, MA 02111-1307, USA.
  */
 
+/* grab gtk_style_get_font and gdk_x11_get_font_name */
+#undef GTK_DISABLE_DEPRECATED
+#undef GDK_DISABLE_DEPRECATED
+#include <gtk/gtkstyle.h>
+#include <gdk/gdkx.h>
+#define GTK_DISABLE_DEPRECATED
+#define GDK_DISABLE_DEPRECATED
+
 #include "terminal-intl.h"
 #include "terminal-accels.h"
 #include "terminal-window.h"
@@ -27,9 +35,9 @@
 #include "terminal.h"
 #include <libgnome/gnome-util.h> /* gnome_util_user_shell */
 #include <libgnome/gnome-url.h> /* gnome_url_show */
-#include <gdk/gdkx.h>
 #include <string.h>
 #include <stdlib.h>
+#include <unistd.h>
 
 struct _TerminalScreenPrivate
 {
@@ -46,6 +54,9 @@ struct _TerminalScreenPrivate
   char **override_command;
   GtkWidget *title_editor;
   GtkWidget *title_entry;
+  char *working_dir;
+  int child_pid;
+  guint recheck_working_dir_idle;
 };
 
 static GList* used_ids = NULL;
@@ -79,10 +90,23 @@ static void terminal_screen_widget_child_died        (GtkWidget      *term,
 static void terminal_screen_widget_selection_changed (GtkWidget      *term,
                                                       TerminalScreen *screen);
 
+static void terminal_screen_setup_dnd                (TerminalScreen *screen);
+
 static void reread_profile (TerminalScreen *screen);
 
 static void rebuild_title  (TerminalScreen *screen);
 
+static void queue_recheck_working_dir (TerminalScreen *screen);
+
+static gboolean xfont_is_monospace                   (const char *fontname);
+static char*    make_xfont_monospace                 (const char *fontname);
+static char*    make_xfont_char_cell                 (const char *fontname);
+static char*    make_xfont_have_size_from_other_font (const char *fontname,
+                                                     const char *other_font);
+static PangoFontDescription* make_font_monospace (const PangoFontDescription *font);
+
+static GdkFont* load_fonset_without_error (const gchar *fontset_name);
+
 static gpointer parent_class;
 static guint signals[LAST_SIGNAL] = { 0 };
 
@@ -118,11 +142,20 @@ terminal_screen_get_type (void)
 
 static void
 terminal_screen_init (TerminalScreen *screen)
-{  
+{
+  char buf[PATH_MAX+1];
+  
   screen->priv = g_new0 (TerminalScreenPrivate, 1);
+
+  screen->priv->working_dir = g_strdup (getcwd (buf, sizeof (buf)));
+  if (screen->priv->working_dir == NULL) /* shouldn't ever happen */
+    screen->priv->working_dir = g_strdup (g_get_home_dir ());
+  screen->priv->child_pid = -1;
+
+  screen->priv->recheck_working_dir_idle = 0;
   
   screen->priv->term = terminal_widget_new ();
-
+  
   g_object_ref (G_OBJECT (screen->priv->term));
   gtk_object_sink (GTK_OBJECT (screen->priv->term));
 
@@ -131,6 +164,8 @@ terminal_screen_init (TerminalScreen *screen)
   
   terminal_widget_match_add (screen->priv->term,
                              "(((news|telnet|nttp|file|http|ftp|https)://)|(www|ftp)[-A-Za-z0-9]*\\.)[-A-Za-z0-9\\.]+(:[0-9]*)?/[-A-Za-z0-9_\\$\\.\\+\\!\\*\\(\\),;:@&=\\?/~\\#\\%]*[^]'\\.}>\\) ,\\\"]");
+
+  terminal_screen_setup_dnd (screen);
   
   g_object_set_data (G_OBJECT (screen->priv->term),
                      "terminal-screen",
@@ -219,10 +254,14 @@ terminal_screen_finalize (GObject *object)
   
   g_object_unref (G_OBJECT (screen->priv->term));
 
+  if (screen->priv->recheck_working_dir_idle)
+    g_source_remove (screen->priv->recheck_working_dir_idle);
+  
   g_free (screen->priv->raw_title);
   g_free (screen->priv->cooked_title);
   g_free (screen->priv->matched_string);
   g_strfreev (screen->priv->override_command);
+  g_free (screen->priv->working_dir);
   
   g_free (screen->priv);
   
@@ -432,10 +471,18 @@ update_color_scheme (TerminalScreen *screen)
     return;
 
   terminal_profile_get_palette (screen->priv->profile,
-                                palette);  
+                                palette);
 
-  terminal_profile_get_color_scheme (screen->priv->profile,
-                                     &fg, &bg);
+  if (terminal_profile_get_use_theme_colors (screen->priv->profile))
+    {
+      fg = screen->priv->term->style->text[GTK_STATE_NORMAL];
+      bg = screen->priv->term->style->base[GTK_STATE_NORMAL];
+    }
+  else
+    {
+      terminal_profile_get_color_scheme (screen->priv->profile,
+                                         &fg, &bg);
+    }
 
   terminal_widget_set_colors (screen->priv->term,
                               &fg, &bg, palette);
@@ -446,25 +493,125 @@ terminal_screen_update_on_realize (GtkWidget      *term,
                                    TerminalScreen *screen)
 {
   TerminalProfile *profile;
-  GdkFont *font;
   
   profile = screen->priv->profile;
   
   update_color_scheme (screen);
-  
-  font = gdk_fontset_load (terminal_profile_get_x_font (profile));
-  if (font == NULL)
-    {      
-      g_printerr (_("Could not load font \"%s\"\n"),
-                  terminal_profile_get_x_font (profile));
+
+  if (terminal_widget_supports_pango_fonts ())
+    {
+      if (terminal_profile_get_use_system_font (profile))
+        {
+          PangoFontDescription *desc;
+
+          /* Try creating monospace variant of system font */
+          desc = make_font_monospace (term->style->font_desc);
+
+          if (desc)
+            {
+              terminal_widget_set_pango_font (term, desc);
+              
+              pango_font_description_free (desc);
+            }
+          else
+            {
+              /* At least use the same size */
+              desc = pango_font_description_copy (terminal_profile_get_font (profile));
+
+              if (pango_font_description_get_set_fields (term->style->font_desc) &
+                  PANGO_FONT_MASK_SIZE)
+                {
+                  pango_font_description_set_size (desc,
+                                                   pango_font_description_get_size (term->style->font_desc));
+                }
+
+              terminal_widget_set_pango_font (term,
+                                              terminal_profile_get_font (profile));
+
+              pango_font_description_free (desc);
+            }
+        }
+      else
+        {
+          terminal_widget_set_pango_font (term,
+                                          terminal_profile_get_font (profile));
+        }
     }
   else
     {
-      terminal_widget_set_normal_gdk_font (term, font);
-      terminal_widget_set_bold_gdk_font (term, font);
-      gdk_font_unref (font);
-    }
+      GdkFont *font;
 
+      font = NULL;
+      
+       if (terminal_profile_get_use_system_font (screen->priv->profile))
+         {
+            GtkStyle *style;
+            GdkFont *from_desc;
+
+            style = gtk_widget_get_style (GTK_WIDGET(term));
+            from_desc = gdk_font_from_description (style->font_desc);
+            
+            if (from_desc != NULL)
+              {
+                char *font_name;
+                
+                font_name = g_strdup (gdk_x11_font_get_name (from_desc));
+
+                g_assert (font_name);
+                
+                if (!xfont_is_monospace (font_name));
+                {
+                  /* Can't use the system font as-is */
+                  char *fallback;
+
+                  fallback = make_xfont_monospace (font_name);
+
+                  g_assert (fallback);
+                  
+                  font = load_fonset_without_error (fallback);
+
+                  if (font == NULL)
+                    {
+                      g_free (fallback);
+                      fallback = make_xfont_char_cell (font_name);
+                      font = load_fonset_without_error (fallback);
+                    }
+                
+                  if (font == NULL)
+                    {
+                      g_free (fallback);
+                      fallback =
+                        make_xfont_have_size_from_other_font (terminal_profile_get_x_font (profile),
+                                                              font_name);
+                      font = load_fonset_without_error (fallback);
+                    }
+                
+                  g_free (fallback);
+                }
+                
+                g_free (font_name);
+                gdk_font_unref (from_desc);
+              }
+         }
+       
+       if (font == NULL)
+         {
+           font = gdk_fontset_load (terminal_profile_get_x_font (profile));
+           if (font == NULL)
+             {
+               g_printerr (_("Could not load font \"%s\"\n"),
+                           terminal_profile_get_x_font (profile));
+             }
+         }
+
+       if (font)
+         {
+           terminal_widget_set_normal_gdk_font (term, font);
+           terminal_widget_set_bold_gdk_font (term, font);
+           gdk_font_unref (font);
+         }
+    }
+  
   terminal_widget_set_allow_bold (term,
                                   terminal_profile_get_allow_bold (profile));
 
@@ -774,6 +921,8 @@ terminal_screen_launch_child (TerminalScreen *screen)
                                      path,
                                      argv,
                                      env,
+                                     terminal_screen_get_working_dir (screen),
+                                     &screen->priv->child_pid,
                                      &err))
     {
       show_fork_error_dialog (screen, err);
@@ -808,9 +957,9 @@ new_window_callback (GtkWidget      *menu_item,
                      TerminalScreen *screen)
 {
   terminal_app_new_terminal (terminal_app_get (),
-                             screen->priv->profile,
+                             terminal_profile_get_for_new_term (),
                              NULL,
-                             FALSE, FALSE, NULL, NULL);
+                             FALSE, FALSE, NULL, NULL, NULL, NULL);
 }
 
 static void
@@ -818,9 +967,9 @@ new_tab_callback (GtkWidget      *menu_item,
                   TerminalScreen *screen)
 {
   terminal_app_new_terminal (terminal_app_get (),
-                             screen->priv->profile,
+                             terminal_profile_get_for_new_term (),
                              screen->priv->window,
-                             FALSE, FALSE, NULL, NULL);
+                             FALSE, FALSE, NULL, NULL, NULL, NULL);
 }
 
 static void
@@ -1232,10 +1381,12 @@ terminal_screen_button_press_event (GtkWidget      *widget,
     }
 }
 
-static void
-set_raw_title (TerminalScreen *screen,
-               const char     *title)
-{  
+void
+terminal_screen_set_dynamic_title (TerminalScreen *screen,
+                                   const char     *title)
+{
+  g_return_if_fail (TERMINAL_IS_SCREEN (screen));
+  
   if (screen->priv->raw_title && title &&
       strcmp (screen->priv->raw_title, title) == 0)
     return;
@@ -1260,11 +1411,92 @@ set_raw_title (TerminalScreen *screen,
     }
 }
 
+const char*
+terminal_screen_get_dynamic_title (TerminalScreen *screen)
+{
+  g_return_val_if_fail (TERMINAL_IS_SCREEN (screen), NULL);
+  
+  return screen->priv->raw_title;
+}
+
+void
+terminal_screen_set_working_dir (TerminalScreen *screen,
+                                 const char     *dirname)
+{
+  g_return_if_fail (TERMINAL_IS_SCREEN (screen));  
+
+  g_free (screen->priv->working_dir);
+  screen->priv->working_dir = g_strdup (dirname);
+}
+
+const char*
+terminal_screen_get_working_dir (TerminalScreen *screen)
+{
+  g_return_val_if_fail (TERMINAL_IS_SCREEN (screen), NULL);
+
+  /* If we're on Linux and have a child PID, try to update
+   * the working dir
+   */
+  if (screen->priv->child_pid >= 0)
+    {
+      char *file;
+      char buf[PATH_MAX+1];
+      int len;
+      
+      file = g_strdup_printf ("/proc/%d/cwd", screen->priv->child_pid);
+
+      /* Silently ignore failure here, since we may not be on Linux */
+      len = readlink (file, buf, sizeof (buf) - 1);
+
+      if (len >= 0)
+        {
+          buf[len] = '\0';
+          
+          g_free (screen->priv->working_dir);
+          screen->priv->working_dir = g_strdup (buf);
+        }
+      
+      g_free (file);
+    }
+  
+  return screen->priv->working_dir;
+}
+
+static gboolean
+recheck_dir (void *data)
+{
+  TerminalScreen *screen = data;
+
+  screen->priv->recheck_working_dir_idle = 0;
+  
+  /* called just for side effect */
+  terminal_screen_get_working_dir (screen);
+
+  /* remove idle */
+  return FALSE;
+}
+
+static void
+queue_recheck_working_dir (TerminalScreen *screen)
+{
+  if (screen->priv->recheck_working_dir_idle == 0)
+    {
+      screen->priv->recheck_working_dir_idle =
+        g_idle_add_full (G_PRIORITY_LOW + 50,
+                         recheck_dir,
+                         screen,
+                         NULL);
+    }
+}
+
 static void
 terminal_screen_widget_title_changed (GtkWidget      *widget,
                                       TerminalScreen *screen)
-{      
-  set_raw_title (screen, terminal_widget_get_title (widget));
+{
+  terminal_screen_set_dynamic_title (screen,
+                                     terminal_widget_get_title (widget));  
+
+  queue_recheck_working_dir (screen);
 }
 
 static void
@@ -1273,6 +1505,8 @@ terminal_screen_widget_child_died (GtkWidget      *term,
 {
   TerminalExitAction action;
 
+  screen->priv->child_pid = -1;
+  
   action = TERMINAL_EXIT_CLOSE;
   if (screen->priv->profile)
     action = terminal_profile_get_exit_action (screen->priv->profile);
@@ -1303,7 +1537,7 @@ title_entry_changed (GtkWidget      *entry,
 
   text = gtk_editable_get_chars (GTK_EDITABLE (entry), 0, -1);
   
-  set_raw_title (screen, text);
+  terminal_screen_set_dynamic_title (screen, text);
 
   g_free (text);
 }
@@ -1359,6 +1593,7 @@ terminal_screen_edit_title (TerminalScreen *screen,
       
       gtk_entry_set_activates_default (GTK_ENTRY (entry), TRUE);
       gtk_label_set_mnemonic_widget (GTK_LABEL (label), entry);
+      terminal_util_set_labelled_by (entry, GTK_LABEL (label));
       
       gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
       gtk_box_pack_end (GTK_BOX (hbox), entry, TRUE, TRUE, 0);
@@ -1397,3 +1632,558 @@ terminal_screen_edit_title (TerminalScreen *screen,
   gtk_widget_show_all (screen->priv->title_editor);
   gtk_window_present (GTK_WINDOW (screen->priv->title_editor));
 }
+
+enum
+{
+  TARGET_URI_LIST,
+  TARGET_UTF8_STRING,
+  TARGET_TEXT,
+  TARGET_COMPOUND_TEXT,
+  TARGET_STRING,
+  TARGET_COLOR,
+  TARGET_BGIMAGE,
+  TARGET_RESET_BG,
+  TARGET_TEXT_PLAIN,
+  TARGET_MOZ_URL
+};
+
+static void  
+drag_data_received  (GtkWidget *widget, GdkDragContext *context, 
+		     gint x, gint y,
+		     GtkSelectionData *selection_data, guint info,
+		     guint time,
+                     gpointer data)
+{
+  TerminalScreen *screen;
+
+  screen = TERMINAL_SCREEN (data);
+
+#if 0
+  {
+    GList *tmp;
+    char *str;
+    
+    tmp = context->targets;
+    while (tmp != NULL)
+      {
+        GdkAtom atom = GPOINTER_TO_UINT (tmp->data);
+
+        g_print ("Target: %s\n", gdk_atom_name (atom));        
+        
+        tmp = tmp->next;
+      }
+  }
+#endif
+  
+  switch (info)
+    {
+    case TARGET_STRING:
+    case TARGET_UTF8_STRING:
+    case TARGET_COMPOUND_TEXT:
+    case TARGET_TEXT:
+      {
+        char *str;
+        
+        str = gtk_selection_data_get_text (selection_data);
+
+        /* FIXME should this be converted back to locale encoding
+         * before sending to the terminal? Most likely yes.
+         */
+        if (str && *str)
+          {
+            terminal_widget_write_data_to_child (screen->priv->term,
+                                                 str, strlen (str));
+          }
+
+        g_free (str);
+      }
+      break;
+
+    case TARGET_TEXT_PLAIN:
+      {
+        if (selection_data->format != 8 ||
+            selection_data->length == 0)
+          {
+            g_printerr (_("text/plain dropped on terminal had wrong format (%d) or length (%d)\n"),
+                        selection_data->format,
+                        selection_data->length);
+            return;
+          }
+        
+        /* FIXME just brazenly ignoring encoding issues... */
+        terminal_widget_write_data_to_child (screen->priv->term,
+                                             selection_data->data,
+                                             selection_data->length);
+      }
+      break;
+      
+    case TARGET_COLOR:
+      {
+        guint16 *data = (guint16 *)selection_data->data;
+        GdkColor color;
+        GdkColor fg;
+        TerminalProfile *profile;
+
+        if (selection_data->format != 16 ||
+            selection_data->length != 8)
+          {
+            g_printerr (_("Color dropped on terminal had wrong format (%d) or length (%d)\n"),
+                        selection_data->format,
+                        selection_data->length);
+            return;
+          }
+
+        color.red = data[0];
+        color.green = data[1];
+        color.blue = data[2];
+
+        profile = terminal_screen_get_profile (screen);
+
+        if (profile)
+          {
+            terminal_profile_set_background_type (profile,
+                                                  TERMINAL_BACKGROUND_SOLID);
+            terminal_profile_set_use_theme_colors (profile, FALSE);
+            
+            terminal_profile_get_color_scheme (profile,
+                                               &fg, NULL);
+            
+            terminal_profile_set_color_scheme (profile, &fg, &color);
+          }
+      }
+      break;
+
+    case TARGET_MOZ_URL:
+      {
+        GString *str;
+        int i;
+        const guint16 *char_data;
+        int char_len;
+        char *filename;
+        
+        /* MOZ_URL is in UCS-2 but in format 8. BROKEN!
+         *
+         * The data contains the URL, a \n, then the
+         * title of the web page.
+         */
+        if (selection_data->format != 8 ||
+            selection_data->length == 0 ||
+            (selection_data->length % 2) != 0)
+          {
+            g_printerr (_("Mozilla url dropped on terminal had wrong format (%d) or length (%d)\n"),
+                        selection_data->format,
+                        selection_data->length);
+            return;
+          }
+
+        str = g_string_new (NULL);
+        
+        char_len = selection_data->length / 2;
+        char_data = (const guint16*) selection_data->data;
+        i = 0;
+        while (i < char_len)
+          {
+            if (char_data[i] == '\n')
+              break;
+            
+            g_string_append_unichar (str, (gunichar) char_data[i]);
+            
+            ++i;
+          }
+
+        /* drop file:///, else paste in URI literally */
+        filename = g_filename_from_uri (str->str, NULL, NULL);
+        
+        /* FIXME just brazenly ignoring encoding issues, sending
+         * child some UTF-8
+         */
+        if (filename)
+          terminal_widget_write_data_to_child (screen->priv->term,
+                                               filename, strlen (filename));
+        else
+          terminal_widget_write_data_to_child (screen->priv->term,
+                                               str->str,
+                                               str->len);
+
+        g_free (filename);        
+        g_string_free (str, TRUE);
+      }
+      break;
+      
+    case TARGET_URI_LIST:
+      {
+        char *uri_list;
+        char **uris;
+        int i;
+        
+        if (selection_data->format != 8 ||
+            selection_data->length == 0)
+          {
+            g_printerr (_("URI list dropped on terminal had wrong format (%d) or length (%d)\n"),
+                        selection_data->format,
+                        selection_data->length);
+            return;
+          }
+        
+        uri_list = g_strndup (selection_data->data,
+                              selection_data->length);
+
+	uris = g_strsplit (uri_list, "\r\n", 0);
+
+        i = 0;
+        while (uris && uris[i])
+          {
+            char *old;
+            
+            old = uris[i];
+
+            uris[i] = g_filename_from_uri (old, NULL, NULL);
+
+            /* If the URI wasn't a filename, then pass it through
+             * as a URI, so you can DND from Mozilla or whatever
+             */
+            if (uris[i] == NULL)
+              uris[i] = old;
+            else
+              g_free (old);
+            
+            ++i;
+          }
+
+        if (uris)
+          {
+            char *flat;
+            
+            flat = g_strjoinv (" ", uris);
+            terminal_widget_write_data_to_child (screen->priv->term,
+                                                 flat, strlen (flat));
+            g_free (flat);
+          }
+
+        g_strfreev (uris);
+      }
+      break;
+      
+    case TARGET_BGIMAGE:
+      {
+        char *uri_list;
+        char **uris;
+        gboolean exactly_one;
+        
+        if (selection_data->format != 8 ||
+            selection_data->length == 0)
+          {
+            g_printerr (_("Image filename dropped on terminal had wrong format (%d) or length (%d)\n"),
+                        selection_data->format,
+                        selection_data->length);
+            return;
+          }
+        
+        uri_list = g_strndup (selection_data->data,
+                              selection_data->length);
+
+	uris = g_strsplit (uri_list, "\r\n", 0);
+
+	exactly_one = uris[0] != NULL && (uris[1] == NULL || uris[1][0] == '\0');
+
+        if (exactly_one)
+          {
+            TerminalProfile *profile;
+            char *filename;
+            GError *err;
+
+            err = NULL;
+            filename = g_filename_from_uri (uris[0],
+                                            NULL,
+                                            &err);
+
+            if (err)
+              {
+                g_printerr (_("Error converting URI \"%s\" into filename: %s\n"),
+                            uris[0], err->message);
+
+                g_error_free (err);
+              }
+
+            profile = terminal_screen_get_profile (screen);
+            
+            if (filename && profile)
+              {
+                terminal_profile_set_background_type (profile,
+                                                      TERMINAL_BACKGROUND_IMAGE);
+                
+                terminal_profile_set_background_image_file (profile,
+                                                            filename);
+              }
+
+            g_free (filename);
+          }
+
+        g_strfreev (uris);
+        g_free (uri_list);
+      }
+      break;
+    case TARGET_RESET_BG:
+      {
+        TerminalProfile *profile;
+
+        profile = terminal_screen_get_profile (screen);
+        
+        if (profile)
+          {
+            terminal_profile_set_background_type (profile,
+                                                  TERMINAL_BACKGROUND_SOLID);
+          }
+      }
+      break;
+    }
+}
+
+static void
+terminal_screen_setup_dnd (TerminalScreen *screen)
+{
+  static GtkTargetEntry target_table[] = {
+    { "application/x-color", 0, TARGET_COLOR },
+    { "property/bgimage",    0, TARGET_BGIMAGE },
+    { "x-special/gnome-reset-background", 0, TARGET_RESET_BG },
+    { "text/uri-list",  0, TARGET_URI_LIST },
+    { "text/x-moz-url",  0, TARGET_MOZ_URL },
+    { "UTF8_STRING", 0, TARGET_UTF8_STRING },
+    { "COMPOUND_TEXT", 0, TARGET_COMPOUND_TEXT },
+    { "TEXT", 0, TARGET_TEXT },
+    { "STRING",     0, TARGET_STRING },
+    /* text/plain problematic, we don't know its encoding */
+    { "text/plain", 0, TARGET_TEXT_PLAIN }
+    /* add when gtk supports it perhaps */
+    /* { "text/unicode", 0, TARGET_TEXT_UNICODE } */
+  };
+  
+  g_signal_connect (G_OBJECT (screen->priv->term), "drag_data_received",
+                    G_CALLBACK (drag_data_received), screen);
+  
+  gtk_drag_dest_set (GTK_WIDGET (screen->priv->term),
+                     GTK_DEST_DEFAULT_MOTION |
+                     GTK_DEST_DEFAULT_HIGHLIGHT |
+                     GTK_DEST_DEFAULT_DROP,
+                     target_table, G_N_ELEMENTS (target_table),
+                     GDK_ACTION_COPY);
+}
+
+/* Indices of some xlfd string entries.  The first entry is 0 */
+#define XLFD_WEIGHT_INDEX			2
+#define XLFD_SLANT_INDEX			3
+#define XLFD_SIZE_IN_PIXELS_INDEX		6
+#define XLFD_SIZE_IN_POINTS_INDEX		7
+#define XLFD_HORIZONTAL_RESOLUTION_INDEX	8
+#define XLFD_VERTICAL_RESOLUTION_INDEX		9
+#define XLFD_SPACING_INDEX		        10
+
+#define XLFD_N_FIELDS 14
+
+static char*
+xlfd_get_nth_field (const char *xlfd,
+                    int         n)
+{
+  char **split;
+  char *ret;
+  int i;
+  
+  /* Remember fontsets with the appended ",-*-" and stuff
+   * like that...
+   */
+  
+  split = g_strsplit (xlfd, "-", XLFD_N_FIELDS);
+
+  i = 0;
+  while (i < n)
+    {
+      if (split[i] == NULL)
+        return NULL;
+
+      ++i;
+    }
+  
+  ret = g_strdup (split[n]);
+  
+  g_strfreev (split);
+
+  return ret;
+}
+
+static char*
+xlfd_replace_nth_field (const char *xlfd,
+                        int         n,
+                        const char *new_value)
+{
+  char **split;
+  char *ret;
+  int i;
+  
+  /* Remember fontsets with the appended ",-*-" and stuff
+   * like that...
+   */
+  
+  split = g_strsplit (xlfd, "-", XLFD_N_FIELDS);
+
+  i = 0;
+  while (i < n)
+    {
+      if (split[i] == NULL)
+        return NULL;
+
+      ++i;
+    }
+
+  g_free (split[n]);
+  split[n] = g_strdup (new_value);
+  
+  ret = g_strjoinv ("-", split);
+  
+  g_strfreev (split);
+
+  return ret;
+}
+
+static gboolean
+xfont_is_monospace (const char *fontname)
+{
+  char *spacing;
+  gboolean ret;
+  
+  spacing = xlfd_get_nth_field (fontname, XLFD_SPACING_INDEX);
+
+  if (spacing == NULL)
+    return FALSE;
+
+  ret = (*spacing == 'm' || *spacing == 'c');
+
+  g_free (spacing);
+
+  return ret;
+}
+
+static char*
+make_xfont_monospace (const char *fontname)
+{
+  char *ret;
+  
+  ret = xlfd_replace_nth_field (fontname, XLFD_SPACING_INDEX, "m");
+  if (ret == NULL)
+    ret = g_strdup (fontname);
+
+  return ret;
+}
+
+static char*
+make_xfont_char_cell (const char *fontname)
+{
+  char *ret;
+  
+  ret = xlfd_replace_nth_field (fontname, XLFD_SPACING_INDEX, "c");
+  if (ret == NULL)
+    ret = g_strdup (fontname);
+
+  return ret;
+}
+
+static char*
+make_xfont_have_size_from_other_font (const char *fontname,
+                                      const char *other_font)
+{
+  char *size_pixels;
+  char *size_points;
+  char *ret;
+
+  ret = NULL;
+  size_pixels = xlfd_get_nth_field (other_font, XLFD_SIZE_IN_PIXELS_INDEX);
+  size_points = xlfd_get_nth_field (other_font, XLFD_SIZE_IN_PIXELS_INDEX);
+
+  if (size_pixels && size_points)
+    {
+      ret = xlfd_replace_nth_field (fontname, XLFD_SIZE_IN_PIXELS_INDEX,
+                                    size_pixels);
+      if (ret)
+        {
+          char *tmp;
+
+          tmp = xlfd_replace_nth_field (ret, XLFD_SIZE_IN_POINTS_INDEX,
+                                        size_points);
+
+          if (tmp)
+            {
+              g_free (ret);
+              ret = tmp;
+            }
+        }
+    }
+
+  g_free (size_pixels);
+  g_free (size_points);
+
+  return ret;
+}
+
+static PangoFontDescription*
+make_font_monospace (const PangoFontDescription *font)
+{
+  PangoFontDescription *ret;
+
+  return NULL;
+  
+  ret = pango_font_description_copy (font);
+
+  /* FIXME use a heuristic here such as getting width of "i" and
+   * "M" and if they match our font is monospace.
+   */
+  
+  return ret;
+}
+
+static GdkFont*
+load_fonset_without_error (const gchar *fontset_name)
+{
+  XFontSet fontset;
+  int  missing_charset_count;
+  char **missing_charset_list;
+  char *def_string;
+  GdkFont *ret;
+  
+  fontset = XCreateFontSet (GDK_DISPLAY (), fontset_name,
+			    &missing_charset_list, &missing_charset_count,
+			    &def_string);
+
+  if (missing_charset_count)
+    {
+      /* The whole point of this function is not to print the below stuff,
+       * it's just here because I might change the function to optionally
+       * return these things in a GError
+       */
+      
+      int i;
+      GString *str;
+
+      str = g_string_new (NULL);
+      g_string_append_printf (str,
+                              "The font \"%s\" does not support all the required character sets for the current locale \"%s\"\n",
+                              fontset_name, setlocale (LC_ALL, NULL));
+      for (i=0;i<missing_charset_count;i++)
+	g_string_append_printf (str,
+                                "  (Missing character set \"%s\")\n",
+                                missing_charset_list[i]);
+      XFreeStringList (missing_charset_list);
+      
+      /* g_printerr ("%s", str->str); */
+
+      g_string_free (str, TRUE);
+    }
+
+  if (fontset == NULL)
+    return NULL;
+  
+  /* This should succeed, and be fast since we still have "fontset" open */     
+  ret = gdk_fontset_load (fontset_name);
+
+  XFreeFontSet (GDK_DISPLAY (), fontset);
+  
+  return ret;
+}
